/**
 * Copyright (C) 2001 Billy Biggs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <screen.h>
#include <signal.h>
#include <slang.h>
#include <assert.h>
#include <song.h>
#include <ctype.h>
#include <unistd.h>
#include <utils.h>
#include <patbar.h>
#include <songbar.h>

using namespace std;

static volatile int Screen_Size_Changed = 0;
static const int numbars = 8;
static char inputbuf[ 80 ];
static const int SL_KEY_F10 = 0x1000;
static const int SL_KEY_F11 = 0x1001;
static const int SL_KEY_F12 = 0x1002;
static const int SL_KEY_SHIFTLEFT = 0x1003;
static const int SL_KEY_SHIFTRIGHT = 0x1004;

static void inputBuffer( char *prompt )
{
	int i = 0;
	unsigned int c;
	SLsmg_gotorc( SLtt_Screen_Rows - 1, 0 );
        SLsmg_erase_eol();
        SLsmg_write_string( prompt );
	SLsmg_refresh();

	for(;;) {
		if( SLang_input_pending( 0 ) ) {
			c = SLkp_getkey();
			if( c == '\r' ) { inputbuf[ i ] = '\0'; break; }
			if( c == SL_KEY_BACKSPACE ) {
				SLsmg_gotorc( SLtt_Screen_Rows - 1, SLsmg_get_column() - 1 );
				SLsmg_write_char( ' ' );
				SLsmg_gotorc( SLtt_Screen_Rows - 1, SLsmg_get_column() - 1 );
				SLsmg_refresh();
				i--;
				if( i == -1 ) i = 0;
				continue;
			}
			inputbuf[ i++ ] = c;
			SLsmg_write_char( c );
			SLsmg_refresh();
		}
		usleep( 1 );
	}
	SLsmg_gotorc( SLtt_Screen_Rows - 1, 0 );
        SLsmg_erase_eol();
}

static void sigwinch_handler( int sig )
{
	Screen_Size_Changed = 1;
	SLsignal( SIGWINCH, sigwinch_handler );
}

Screen::Screen( void )
	: pos( 0 ), octave( 3 ), intsync( true ), patmode( false ), curcol( 0 ), copyPattern( 0 )
{
	int ret, i;

	// Initialize s-lang stuff.
	SLsignal( SIGWINCH, sigwinch_handler );
	SLtt_get_terminfo();
	SLang_init_tty( 0, 1, 0 );
	ret = SLkp_init();
	assert( ret != -1 );
	ret = SLsmg_init_smg();
	assert( ret != -1 );

	// Define all the extra keys.
	SLkp_define_keysym( "[21~", SL_KEY_F10 );
	SLkp_define_keysym( "[23~", SL_KEY_F11 );
	SLkp_define_keysym( "[24~", SL_KEY_F12 );
	SLkp_define_keysym( "[2D", SL_KEY_SHIFTLEFT );
	SLkp_define_keysym( "[2C", SL_KEY_SHIFTRIGHT );

	// Cursor colour.
	SLtt_set_color( Screen::CursorColour, 0, "white", "blue" );

	// Hilight colour.
	SLtt_set_color( Screen::HilightColour, 0, "green", "black" );

	// Hilight + cursor colour.
	SLtt_set_color( Screen::HilightCursorColour, 0, "green", "blue" );

	// Muted channel colour.
	SLtt_set_color( Screen::MutedChannelColour, 0, "blue", "black" );

	// Realtime Light colour.
	SLtt_set_color( Screen::RealtimeLightColour, 0, "white", "red" );

	// Play position colour.
	SLtt_set_color( Screen::PlayPositionColour, 0, "red", "black" );

	// Loop hilight colour.
	SLtt_set_color( Screen::LoopHilightColour, 0, "black", "green" );

	// Loop inactive hilight colour.
	SLtt_set_color( Screen::LoopInactiveColour, 0, "black", "red" );

	// Default colour.
	SLtt_set_color( 0, 0, "white", "black" );

	// Remember to stay at the default colour from now on.
	SLsmg_set_color( Screen::DefaultColour );

	// Create the bars array.
	bars = new ChanBar*[ numbars ];

	// Create the pattern bars.
	patbars = new PatBar*[ numbars ];
	for( i = 0; i < numbars; i++ ) {
		patbars[ i ] = new PatBar( i, 3, 5 + ( i * 9 ) );
	}

	// Create the channel bars.
	songbars = new SongBar*[ numbars ];
	for( i = 0; i < numbars; i++ ) {
		songbars[ i ] = new SongBar( i, 3, 5 + ( i * 9 ) );
	}

	// Clear the virtual screen.
	SLsmg_cls();

	// Print the help.  This is only printed once.
	printHelp();

	// Switch the mode to setup everything initially.
	switchMode();

	// Reset the cursor.
	patbars[ 0 ]->setCursor( 0 );
	songbars[ 0 ]->setCursor( 0 );
	moveCursorPosition( 0, 0 );

	// Refresh the screen.
	refresh();
}

Screen::~Screen( void )
{
	SLsmg_reset_smg();
	SLang_reset_tty();
}

void Screen::switchMode( void )
{
	int i;

	patmode = !patmode;
	if( patmode ) {
		for( i = 0; i < numbars; ++i ) {
			bars[ i ] = patbars[ i ];
			bars[ i ]->refresh();
		}
		curstate = patstate;
	} else {
		for( i = 0; i < numbars; ++i ) {
			bars[ i ] = songbars[ i ];
			bars[ i ]->refresh();
		}
		curstate = songstate;
	}
	moveCursorPosition( curstate.cursorrow, curcol );
}

void Screen::refresh( void ) const
{
	// Need to refresh the header.
	printHeader();

	// Need to refresh the sidebar too.
	printSidebar();

	// Print beat pulse.
	if( midiControl.beatPulse() ) {
		SLsmg_gotorc( 0, SLtt_Screen_Cols - 2 );
		SLsmg_printf( "*" );
	} else {
		SLsmg_gotorc( 0, SLtt_Screen_Cols - 2 );
		SLsmg_printf( " " );
	}

	// Announce if we got leet root access.
	SLsmg_gotorc( SLtt_Screen_Rows - 2, 0 );
	if( midiControl.isRealtime() ) {
		SLsmg_forward( SLtt_Screen_Cols - 2 );
		SLsmg_set_color( Screen::RealtimeLightColour );
		SLsmg_printf( "RT" );
		SLsmg_set_color( Screen::DefaultColour );
	}

	// Always leave the cursor in the bottom-right corner.
	SLsmg_gotorc( SLtt_Screen_Rows - 1, SLtt_Screen_Cols - 1 );
	SLsmg_refresh();
}

bool Screen::pollScreen( void )
{
	unsigned int c;

	// Handle a screen size change (annoying).
	if( Screen_Size_Changed ) {
		SLtt_get_screen_size();
		SLsmg_reinit_smg();
		Screen_Size_Changed = 0;
		refresh();
		printHelp();
		for( int i = 0; i < numbars; i++ ) {
			bars[ i ]->refresh();
		}
	}

	// This input call will throttle the main loop.
	if( SLang_input_pending( 0 ) ) {
		c = SLkp_getkey();
		c = toupper( c );
		switch( c ) {

		case 17: return true; // Ctrl-Q is quit

		// Channel mutes.
		case SL_KEY_F(1):	bars[ 0 ]->toggleChannelMute(); break;
		case SL_KEY_F(2):	bars[ 1 ]->toggleChannelMute(); break;
		case SL_KEY_F(3):	bars[ 2 ]->toggleChannelMute(); break;
		case SL_KEY_F(4):	bars[ 3 ]->toggleChannelMute(); break;
		case SL_KEY_F(5):	bars[ 4 ]->toggleChannelMute(); break;
		case SL_KEY_F(6):	bars[ 5 ]->toggleChannelMute(); break;
		case SL_KEY_F(7):	bars[ 6 ]->toggleChannelMute(); break;
		case SL_KEY_F(8):	bars[ 7 ]->toggleChannelMute(); break;

		// Song Control.
		case SL_KEY_F(9):  ttrkSong.start(); break;
		case SL_KEY_F10:   ttrkSong.rewind(); ttrkSong.start(); break;
		case SL_KEY_F11:   ttrkSong.stop(); break;
		case SL_KEY_F12:   ttrkSong.rewind(); break;

		case SL_KEY_DOWN:
			moveCursorPosition( curstate.cursorrow + 1, curcol ); break;
		case SL_KEY_UP:
			moveCursorPosition( curstate.cursorrow - 1, curcol ); break;
		case SL_KEY_LEFT:
			moveCursorPosition( curstate.cursorrow, curcol - 1 ); break;
		case SL_KEY_RIGHT:
			moveCursorPosition( curstate.cursorrow, curcol + 1 ); break;

		case SL_KEY_SHIFTLEFT:
			moveCursorPosition( curstate.cursorrow, curcol - numbars ); break;
		case SL_KEY_SHIFTRIGHT:
			moveCursorPosition( curstate.cursorrow, curcol + numbars ); break;

		case SL_KEY_IC:     bars[ curcol ]->editMidiChannel( 1 ); break;
		case SL_KEY_DELETE: bars[ curcol ]->editMidiChannel( -1 ); break;

		case '/': bars[ curcol ]->editChannelVolume( -1 ); break;
		case '*': bars[ curcol ]->editChannelVolume( 1 ); break;

		case SL_KEY_HOME:
			if( patmode ) patbars[ curcol ]->editCursorVolume( 1 ); break;
		case SL_KEY_END:
			if( patmode ) patbars[ curcol ]->editCursorVolume( -1 ); break;

		case '[': bars[ curcol ]->editPattern( -1 ); break;
		case ']': bars[ curcol ]->editPattern( 1 ); break;

		case '+': ++octave; break;
		case '-': --octave; break;

		case '\'': if( patmode ) patbars[ curcol ]->transposeNotes( -1 ); break;
		case '\\': if( patmode ) patbars[ curcol ]->transposeNotes( 1 ); break;

		case SL_KEY_PPAGE:
			if( patmode ) {
				setPosition( pos - 1 );
			} else {
				moveCursorPosition( curstate.cursorrow - 8, curcol );
			}
			break;
		case SL_KEY_NPAGE:
			if( patmode ) {
				setPosition( pos + 1 );
			} else {
				moveCursorPosition( curstate.cursorrow + 8, curcol );
			}
			break;

		case '`':
			if( patmode ) {
			 	patbars[ curcol ]->toggleCursorSlide();
				moveCursorPosition( patstate.cursorrow + 1, curcol );
			}
			break;

		case ' ':
			if( patmode ) {
				setNote( Song::NoteEmpty );
			} else {
				bars[ curcol ]->editPattern( 0 - ( SongMaxPat + 1 ) );
				moveCursorPosition( curstate.cursorrow + 1, curcol );
			}
			break;
		case 'Z': setNote( (octave + 1) * 12 + 0 ); break;
		case 'S': setNote( (octave + 1) * 12 + 1 ); break;
		case 'X': setNote( (octave + 1) * 12 + 2 ); break;
		case 'D': setNote( (octave + 1) * 12 + 3 ); break;
		case 'C': setNote( (octave + 1) * 12 + 4 ); break;
		case 'V': setNote( (octave + 1) * 12 + 5 ); break;
		case 'G': setNote( (octave + 1) * 12 + 6 ); break;
		case 'B': setNote( (octave + 1) * 12 + 7 ); break;
		case 'H': setNote( (octave + 1) * 12 + 8 ); break;
		case 'N': setNote( (octave + 1) * 12 + 9 ); break;
		case 'J': setNote( (octave + 1) * 12 + 10 ); break;
		case 'M': setNote( (octave + 1) * 12 + 11 ); break;

		case 'Q': setNote( (octave + 1) * 12 + 12 ); break;
		case '2': setNote( (octave + 1) * 12 + 13 ); break;
		case 'W': setNote( (octave + 1) * 12 + 14 ); break;
		case '3': setNote( (octave + 1) * 12 + 15 ); break;
		case 'E': setNote( (octave + 1) * 12 + 16 ); break;
		case 'R': setNote( (octave + 1) * 12 + 17 ); break;
		case '5': setNote( (octave + 1) * 12 + 18 ); break;
		case 'T': setNote( (octave + 1) * 12 + 19 ); break;
		case '6': setNote( (octave + 1) * 12 + 20 ); break;
		case 'Y': setNote( (octave + 1) * 12 + 21 ); break;
		case '7': setNote( (octave + 1) * 12 + 22 ); break;
		case 'U': setNote( (octave + 1) * 12 + 23 ); break;

		case 'I': setNote( (octave + 1) * 12 + 24 ); break;
		case '9': setNote( (octave + 1) * 12 + 25 ); break;
		case 'O': setNote( (octave + 1) * 12 + 26 ); break;
		case '0': setNote( (octave + 1) * 12 + 27 ); break;
		case 'P': setNote( (octave + 1) * 12 + 28 ); break;

		case SL_KEY_BACKSPACE: setNote( Song::NoteOff ); break;

		case '': inputTempo(); break;
		case '': ttrkSong.setLoopPosition( pos + 1 ); break;
		case '': ttrkSong.setLoopDestination( pos ); break;
		case '': ttrkSong.setStartPos( pos ); break;
		case  19 : saveSong(); break; // Ctrl-S
		case '': loadSong(); break; // Ctrl-S
		case '': toggleSync(); break;
		case '': saveCopy(); break;
		case '': pasteCopy(); break;
		case '': enterChanName(); break;
		case '': enterPatName(); break;
		case  '!': ttrkSong.toggleLoop(); break;
		case '	': switchMode(); break;
		default: SLtt_beep(); break;
		}
	}

	refresh();
	return false;
}

void Screen::printHeader( void ) const
{
	char *status;
	char *loopstat;
	char tempo[ 4 ];

	if( ttrkSong.isPlaying() ) {
		status = "Playing";
	} else {
		status = "Stopped";
	}

	if( ttrkSong.isLoopActive() ) {
		loopstat = "  Active";
	} else {
		loopstat = "Inactive";
	}

	if( intsync ) {
		sprintf( tempo, "%03d", ttrkSong.getTempo() );
	} else {
		strcpy( tempo, "---" );
	}

	SLsmg_gotorc( 0, 0 );
	SLsmg_printf( " %03x:%02d %s          Speed: %sbpm            Loop: %s, %03x -> %03x",
		ttrkSong.getCurPat(), ttrkSong.getCurBeat(), status, tempo, loopstat,
		ttrkSong.getLoopPosition(), ttrkSong.getLoopDestination() );
	SLsmg_gotorc( 1, 0 );
	SLsmg_printf( "[%03x]                  Viewing: %03x             Octave: %02d",
		ttrkSong.getStartPos(), pos, octave );
}

void Screen::printSidebar( void ) const
{
	int i;

	SLsmg_gotorc( 8, 0 );
	SLsmg_printf( "====" );

	int scrsize = SLtt_Screen_Rows - 3 - Screen::HeaderSize - Screen::BottomSize;
	if( patmode ) {
		for( i = patstate.scrolltop; i < SongPatLen; i++ ) {
			if( i >= ( patstate.scrolltop + scrsize ) ) break;
			SLsmg_gotorc( 9 + i - patstate.scrolltop, 0 );
			if( ttrkSong.getCurPat() == pos && ttrkSong.getCurBeat() == i ) {
				SLsmg_set_color( Screen::PlayPositionColour );
			} else if( i % 4 == 0 ) SLsmg_set_color( Screen::HilightColour );
			SLsmg_printf( " %02d: ", i );
			SLsmg_set_color( Screen::DefaultColour );
		}
	} else {
		for( i = songstate.scrolltop; i < SongLen; i++ ) {
			if( i >= ( songstate.scrolltop + scrsize ) ) break;
			SLsmg_gotorc( 9 + i - songstate.scrolltop, 0 );
			if( ttrkSong.getCurPat() == i ) {
				SLsmg_set_color( Screen::PlayPositionColour );
			} else if( i % 4 == 0 ) SLsmg_set_color( Screen::HilightColour );
			SLsmg_printf( "%03x:", i );
			if( i < ttrkSong.getLoopPosition() && i >= ttrkSong.getLoopDestination() ) {
				if( ttrkSong.isLoopActive() ) {
					SLsmg_set_color( Screen::LoopHilightColour );
				} else {
					SLsmg_set_color( Screen::LoopInactiveColour );
				}
			}
			SLsmg_printf( " " );
			SLsmg_set_color( Screen::DefaultColour );
		}
	}
}

void Screen::printHelp( void ) const
{
	int currow = SLtt_Screen_Rows - 8;
	SLsmg_gotorc( currow, 0 );
	SLsmg_printf( "%13s %20s %20s %20s",
		"Quit: ^Q", "SetStart: ^F   ", "View: PgUp/Dn", "ChanMute: F1-F8" );
	SLsmg_gotorc( currow + 1, 0 );
	SLsmg_printf( "%13s %20s %20s %20s",
		"Load: ^A", "LoopStart: ^L   ", "Octave: +/-    ", "Start: F9   " );
	SLsmg_gotorc( currow + 2, 0 );
	SLsmg_printf( "%13s %20s %20s %20s",
		"Save: ^S", "LoopDest: ^K   ", " NoteVel: Hm/End ", "Rew+Start: F10  " );
	SLsmg_gotorc( currow + 3, 0 );
	SLsmg_printf( "%13s %20s %20s %20s",
		"Tempo: ^T", "ExtSync: ^E   ", "MidiCh: Ins/Del", "Stop: F11  " );
	SLsmg_gotorc( currow + 4, 0 );
	SLsmg_printf( "%13s %20s %20s %20s",
		"ChanName: ^N", "Copy: ^C   ", "ChanVol: *//    ", "Rewind: F12  " );

	SLsmg_gotorc( currow + 5, 0 );
	SLsmg_printf( "%13s %20s %20s %20s",
		" PatName: ^P", "Paste: ^V   ", "Pattern: [/]    ", "View: <tab>" );
}

void Screen::showStatusMessage( char *newmsg ) const
{
	SLsmg_gotorc( SLtt_Screen_Rows - 1, 0 );
        SLsmg_erase_eol();
        SLsmg_write_string( newmsg );
	SLsmg_gotorc( SLtt_Screen_Rows - 1, SLtt_Screen_Cols - 1 );
	SLsmg_refresh();
}

void Screen::moveCursorPosition( int newrow, int newcol )
{
	int i;

	// Wrap up-down.
	if( patmode ) {
		if( newrow < 0 ) newrow = 31;
		newrow %= SongPatLen;
	} else {
		if( newrow < 0 ) newrow = 0;
		if( newrow >= SongLen ) newrow = SongLen - 1;
	}

	// Wrap left-right.
	newcol += bars[ 0 ]->getIndex();
	if( newcol < 0 ) newcol = 0;
	if( newcol >= SongMaxChan ) newcol = SongMaxChan - 1;

	int range = ( newcol / numbars );
	for( i = 0; i < numbars; i++ ) {
		patbars[ i ]->setChannel( (numbars * range ) + i );
		songbars[ i ]->setChannel( (numbars * range ) + i );
	}
	newcol %= numbars;

	patbars[ curcol ]->setCursor( -1 );
	songbars[ curcol ]->setCursor( -1 );
	curstate.scrolltop = bars[ newcol ]->setCursor( newrow );
	for( i = 0; i < numbars; i++ ) {
		bars[ i ]->setScrollTop( curstate.scrolltop );
	}
	printSidebar();
	curstate.cursorrow = newrow;
	curcol = newcol;
	if( patmode ) {
		patstate = curstate;
	} else {
		songstate = curstate;
		setPosition( songstate.cursorrow );
	}
	printSidebar();
}

void Screen::setPosition( int newpos )
{
	if( newpos < 0 ) newpos = 0;
	if( newpos > SongLen ) newpos = SongLen;
	pos = newpos;

	for( int i = 0; i < numbars; i++ ) {
		patbars[ i ]->setPosition( pos );
		if( patmode ) patbars[ i ]->refresh();
	}
	songstate.cursorrow = pos;

	refresh();
}

void Screen::inputTempo( void ) const
{
	inputBuffer( "Enter Tempo: " );
	ttrkSong.setTempo( atoi( inputbuf ) );
	refresh();
}

void Screen::saveSong( void ) const
{
	inputBuffer( "Enter Save Filename: " );
	if( inputbuf[ 0 ] == '\0' ) {
		showStatusMessage( "Save cancelled." );
		return;
	}
	string result = expand_path( string( inputbuf ) );
	if( !ttrkSong.saveSong( result.c_str() ) ) {
		showStatusMessage( "Save failed." );
	}
	refresh();
	showStatusMessage( "Saved." );
}

void Screen::loadSong( void ) const
{
	inputBuffer( "Enter Load Filename: " );
	if( inputbuf[ 0 ] == '\0' ) {
		showStatusMessage( "Load cancelled." );
		return;
	}
	string result = expand_path( string( inputbuf ) );
	if( !ttrkSong.loadSong( result.c_str() ) ) {
		showStatusMessage( "Load failed." );
	}
	for( int i = 0; i < numbars; i++ ) {
		bars[ i ]->refresh();
		showStatusMessage( "Song loaded." );
	}
	refresh();
}

void Screen::enterPatName( void ) const
{
	if( !bars[ curcol ]->patternSet() ) {
		showStatusMessage( "No pattern active on current channel." );
		return;
	}
	inputBuffer( "Enter Pattern Name: " );
	// For now, trim all names to 8 characters.
	inputbuf[ 8 ] = '\0';
	bars[ curcol ]->setPatternName( inputbuf );
	showStatusMessage( "Name set." );
}

void Screen::enterChanName( void ) const
{
	inputBuffer( "Enter Channel Name: " );
	// For now, trim all names to 8 characters.
	inputbuf[ 8 ] = '\0';
	bars[ curcol ]->setChannelName( inputbuf );
	showStatusMessage( "Name set." );
}

void Screen::toggleSync( void )
{
	intsync = !intsync;

	if( intsync ) {
		midiControl.useInternalClock();
		showStatusMessage( "Using internal sync." );
	} else {
		midiControl.useExternalClock();
		showStatusMessage( "Using external sync." );
	}
}

void Screen::saveCopy( void )
{
	if( !patmode ) return;
	SongPattern *pat = ttrkSong.getChannel(
		patbars[ curcol ]->getIndex() )->getSongPattern( pos );
	if( !pat ) return;
	pat->copyTo( &copyPattern );
	showStatusMessage( "Copied." );
}

void Screen::pasteCopy( void )
{
	if( !patmode ) return;
	SongChannel *chan = ttrkSong.getChannel(
		patbars[ curcol ]->getIndex() );
	SongPattern *pat = chan->getSongPattern( pos );
	if( !pat ) {
		for( int i = 0; i < SongMaxPat; i++ ) {
			if( chan->getChannelPattern( i )->isBlank() ) {
				chan->setSongPattern( pos, i + 1 );
				pat = chan->getSongPattern( pos );
				break;
			}
		}
	}
	pat = chan->getSongPattern( pos );
	if( !pat ) return;
	copyPattern.copyTo( pat );
	patbars[ curcol ]->refresh();
	showStatusMessage( "Pasted." );
}

void Screen::setNote( int newnote )
{
	if( patmode ) {
		patbars[ curcol ]->setNote( newnote );
		moveCursorPosition( patstate.cursorrow + 1, curcol );
	}
}

